# vim: ts=4:sts=4:sw=4
#
# @author: <sylvain.herledan@oceandatalab.com>
# @date: 2016-11-03
#
# This file is part of SEAScope, a 3D visualisation and analysis application
# for satellite, in-situ and numerical model data.
#
# Copyright (C) 2014-2023 OceanDataLab
#
# SEAScope is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SEAScope. If not, see <https://www.gnu.org/licenses/>.

"""
This module handles the serialization of granule data objects
"""

import logging
import flatbuffers
import SEAScope.API.GranuleData
import SEAScope.types.data_bucket

logger = logging.getLogger(__name__)


def serialize(builder, data_obj):
    """
    Serialize a granule data using FlatBuffers.

    Parameters
    ----------
    builder : flatbuffers.builder.Builder
        The FlatBuffers builder instance which serializes data. If this
        parameter is None, then a new builder will be created
    data_obj : dict
        Dictionary which contains information about the granule data to
        serialize. Its keys must be the name of the data fields and its values
        must be :obj:`dict` objects that satisfy the requirements of the
        :func:`SEAScope.types.data_bucket.serialize` method


    Returns
    -------
    tuple(flatbuffers.builder.Builder, int)
        A tuple which contains two elements:

        - the :obj:`flatbuffers.builder.Builder` instance which has been used
          to serialize data
        - an :obj:`int` which is the address/offset of the serialized object
          in the builder buffer
    """
    if builder is None:
        builder = flatbuffers.Builder(0)

    cls = SEAScope.API.GranuleData

    # Identify fields stored as unsigned bytes
    uint8_fields = [_ for _ in data_obj.keys()
                    if 'uint8' == data_obj[_]['info']['dataType']]

    _fields = []
    for field in uint8_fields:
        _field = builder.CreateString(field)
        _fields.append(_field)
    fields_len = len(_fields)
    cls.GranuleDataStartFieldsVector(builder, fields_len)
    for _field in _fields:
        builder.PrependUOffsetTRelative(_field)
    fields_vector = builder.EndVector(fields_len)

    _data = []
    for field in uint8_fields:
        _, _d = SEAScope.types.data_bucket.serialize(builder, data_obj[field])
        _data.append(_d)
    cls.GranuleDataStartBucketsVector(builder, fields_len)
    for _d in _data:
        builder.PrependUOffsetTRelative(_d)
    buckets = builder.EndVector(fields_len)

    # Identify fields stored as floats
    float_fields = [_ for _ in data_obj.keys()
                    if 'float32' == data_obj[_]['info']['dataType']]

    _fields = []
    for field in float_fields:
        _field = builder.CreateString(field)
        _fields.append(_field)
    fields_len = len(_fields)
    cls.GranuleDataStartFloatFieldsVector(builder, fields_len)
    for _field in _fields:
        builder.PrependUOffsetTRelative(_field)
    float_fields_vector = builder.EndVector(fields_len)

    _data = []
    for field in float_fields:
        _, _d = SEAScope.types.data_bucket.serialize(builder, data_obj[field])
        _data.append(_d)
    cls.GranuleDataStartBucketsVector(builder, fields_len)
    for _d in _data:
        builder.PrependUOffsetTRelative(_d)
    float_buckets = builder.EndVector(fields_len)

    cls.GranuleDataStart(builder)
    cls.GranuleDataAddFields(builder, fields_vector)
    cls.GranuleDataAddBuckets(builder, buckets)
    cls.GranuleDataAddFloatFields(builder, float_fields_vector)
    cls.GranuleDataAddFloatBuckets(builder, float_buckets)
    g_data = cls.GranuleDataEnd(builder)

    return builder, g_data


def deserialize(o):
    """
    Not implemented
    """
    pass



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: API

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class ListVisibleDataResponse(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsListVisibleDataResponse(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ListVisibleDataResponse()
        x.Init(buf, n + offset)
        return x

    # ListVisibleDataResponse
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ListVisibleDataResponse
    def Uris(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # ListVisibleDataResponse
    def UrisLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # ListVisibleDataResponse
    def UrisIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # ListVisibleDataResponse
    def Ids(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 32
            from SEAScope.API.RenderableId import RenderableId
            obj = RenderableId()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # ListVisibleDataResponse
    def IdsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # ListVisibleDataResponse
    def IdsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

def ListVisibleDataResponseStart(builder): builder.StartObject(2)
def ListVisibleDataResponseAddUris(builder, uris): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(uris), 0)
def ListVisibleDataResponseStartUrisVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def ListVisibleDataResponseAddIds(builder, ids): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(ids), 0)
def ListVisibleDataResponseStartIdsVector(builder, numElems): return builder.StartVector(32, numElems, 8)
def ListVisibleDataResponseEnd(builder): return builder.EndObject()



# vim: set ts=4:sts=4:sw=4
#
# @author: <sylvain.herledan@oceandatalab.com>
#
# This file is part of SEAScope, a 3D visualisation and analysis application
# for satellite, in-situ and numerical model data.
#
# Copyright (C) 2014-2023 OceanDataLab
#
# SEAScope is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SEAScope. If not, see <https://www.gnu.org/licenses/>.

"""
This module provides methods for plotting the data extracted by SEAScope.
"""

import logging

logger = logging.getLogger(__name__)


def plot_2d_data(source_name, field_name, field_data, output_path):
    """Plot data extracted by SEAScope using a polygon.

    Parameters
    ----------
    source_name : str
        Name of the source (granule) data has been extracted from
    field_name : str
        Name of the extracted field
    field_data : numpy.array
        Extracted data (for a single field/channel)
    output_path : str
        Path of the file wherein the plot will be saved
    """
    import matplotlib
    import matplotlib.pyplot
    slice_size = 100
    slices = range(0, len(source_name), slice_size)
    wrapped_srcname = [source_name[i:i+slice_size] for i in slices]

    fig = matplotlib.pyplot.figure(1, figsize=(10, 10), dpi=300)
    ax = fig.add_axes([0, 0, 1, 1], visible=True)
    ax.set_axis_off()
    matplotlib.pyplot.figtext(.5, 1.0, '\n'.join(wrapped_srcname), fontsize=10,
                              ha='center', figure=fig)
    matplotlib.pyplot.figtext(.5, 0.95, field_name, fontsize=16, ha='center',
                              figure=fig)
    _cmap = matplotlib.pyplot.get_cmap('jet')
    ax.imshow(field_data, cmap=_cmap, interpolation='nearest')

    fig.savefig(output_path, bbox_inches='tight')
    matplotlib.pyplot.close()


def plot_transect(dists, transects, output_path):
    """Plot data extracted by SEAScope using a polyline (transect).

    Parameters
    ----------
    dists : :obj:list of :obj:float
        Cumulative distance from origin to each point of the transect, in
        kilometers
    transects: :obj:dict
        Values extracted along the polyline for each (granule, field) pair
    output_path : str
        Path of the file wherein the plot will be saved
    """
    import matplotlib
    import matplotlib.pyplot
    cmap = matplotlib.pyplot.get_cmap('hsv', 1 + len(transects.keys()))

    fig = matplotlib.pyplot.figure(1, figsize=(10, 10), dpi=300)
    ax = fig.add_axes([0, 0, 1, 1], visible=True)
    ax.set_xlabel('Distance (km)')

    k = 0
    for uri in transects:
        slice_size = 120
        slices = range(0, len(uri), slice_size)
        wrapped_uri = [uri[i:i+slice_size] for i in slices]

        ax.plot(dists, transects[uri], c=cmap(k), label='\n'.join(wrapped_uri))
        k = k + 1

    legend = matplotlib.pyplot.legend(bbox_to_anchor=(0., 1.02, 1., .102),
                                      loc=3, ncol=1, mode="expand",
                                      borderaxespad=0.)
    matplotlib.pyplot.setp(legend.get_title(), fontsize='xx-small')

    fig.savefig(output_path, bbox_inches='tight')
    matplotlib.pyplot.close()



# vim: ts=4:sts=4:sw=4
#
# @author: <sylvain.herledan@oceandatalab.com>
# @date: 2016-09-07
#
# This file is part of SEAScope, a 3D visualisation and analysis application
# for satellite, in-situ and numerical model data.
#
# Copyright (C) 2014-2023 OceanDataLab
#
# SEAScope is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SEAScope. If not, see <https://www.gnu.org/licenses/>.

"""
This module handles the serialization and deserialization of variable objects
"""

import logging
import flatbuffers
import SEAScope.API.Variable
import SEAScope.API.RenderingMethod
import SEAScope.types.tag
import SEAScope.types.rendering_cfg

logger = logging.getLogger(__name__)
rendering_methods = SEAScope.types.rendering_cfg.rendering_methods


def serialize(builder, var_obj):
    """
    Serialize a variable using FlatBuffers.

    Parameters
    ----------
    builder : flatbuffers.builder.Builder
        The FlatBuffers builder instance which serializes data. If this
        parameter is None, then a new builder will be created
    var_obj : dict
        Dictionary which contains information about the variable to serialize
        It must have five keys:

        - ``label`` : a :obj:`str` that SEAScope uses to designate the variable

        - ``units`` : a :obj:`str` for the units of the variable

        - ``fields`` : a :obj:`list` of :obj:`str` which contains the name of
          the data fields that must be read in order to recompose the variable.
          For example the U and V components for a vectorfield, or three
          channels for RGB images

        - ``tags`` : a :obj:`dict` which contains tags that describe the
          variable. SEAScope considers a (key, value) pair as a tag.

        - ``defaultRenderingMethod`` : a :obj:`str` that designates the
          rendering method that must be used by default for this variable.
          This parameter must be one of the keys of the
          :const:`SEAScope.types.rendering_cfg.rendering_methods` dictionary


    Returns
    -------
    tuple(flatbuffers.builder.Builder, int)
        A tuple which contains two elements:

        - the :obj:`flatbuffers.builder.Builder` instance which has been used
          to serialize data
        - an :obj:`int` which is the address/offset of the serialized object
          in the builder buffer
    """
    if builder is None:
        builder = flatbuffers.Builder(0)

    _id = int(var_obj['id'])
    _label = builder.CreateString(var_obj['label'])
    _units = builder.CreateString(var_obj['units'])
    fields_buffer = []
    for field in var_obj['fields']:
        _field = builder.CreateString(field)
        fields_buffer.append(_field)
    fields_len = len(fields_buffer)
    SEAScope.API.Variable.VariableStartFieldsVector(builder, fields_len)
    for t in fields_buffer:
        builder.PrependUOffsetTRelative(t)
    _fields = builder.EndVector(fields_len)

    tags_buffer = []
    for k, v in var_obj['tags'].items():
        _, _tag = SEAScope.types.tag.serialize(builder, k, v)
        tags_buffer.append(_tag)
    tags_len = len(tags_buffer)
    SEAScope.API.Variable.VariableStartTagsVector(builder, tags_len)
    for t in tags_buffer:
        builder.PrependUOffsetTRelative(t)
    _tags = builder.EndVector(tags_len)

    rendering = var_obj['defaultRenderingMethod']
    _default_rendering = rendering_methods.get(rendering, None)

    SEAScope.API.Variable.VariableStart(builder)
    SEAScope.API.Variable.VariableAddId(builder, _id)
    SEAScope.API.Variable.VariableAddLabel(builder, _label)
    SEAScope.API.Variable.VariableAddUnits(builder, _units)
    SEAScope.API.Variable.VariableAddFields(builder, _fields)
    SEAScope.API.Variable.VariableAddTags(builder, _tags)
    SEAScope.API.Variable.VariableAddDefaultRenderingMethod(builder,
                                                            _default_rendering)
    variable = SEAScope.API.Variable.VariableEnd(builder)
    return builder, variable


def deserialize(v):
    """
    Rebuild a variable from a FlatBuffers buffer.

    Parameters
    ----------
    buf : bytearray
        The buffer which contains the variable object serialized with
        FlatBuffers

    Returns
    -------
    dict
        The deserialized variable object as a dictionary.
    """
    v_obj = {}
    v_obj['id'] = v.Id()
    v_obj['label'] = v.Label()
    v_obj['units'] = v.Units()
    fields_count = v.FieldsLength()
    v_obj['fields'] = [v.Fields(x) for x in range(0, fields_count)]
    tags_count = v.TagsLength()
    v_obj['tags'] = {}
    for i in range(0, tags_count):
        name, value = SEAScope.types.tag.deserialize(v.Tags(i))
        v_obj['tags'][name] = value

    v_obj['defaultRenderingMethod'] = 'NONE'
    rendering = v.DefaultRenderingMethod()
    for k, v in rendering_methods.items():
        if v == rendering:
            v_obj['defaultRenderingMethod'] = k
            break
    return v_obj



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: API

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class RenderingCfg(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsRenderingCfg(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = RenderingCfg()
        x.Init(buf, n + offset)
        return x

    # RenderingCfg
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # RenderingCfg
    def Rendered(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # RenderingCfg
    def Logscale(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # RenderingCfg
    def Min(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # RenderingCfg
    def Max(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # RenderingCfg
    def Opacity(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # RenderingCfg
    def Zindex(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # RenderingCfg
    def Color(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            x = o + self._tab.Pos
            from SEAScope.API.Color import Color
            obj = Color()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # RenderingCfg
    def Colormap(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # RenderingCfg
    def RenderMethod(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # RenderingCfg
    def ParticlesCount(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # RenderingCfg
    def ParticleTTL(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # RenderingCfg
    def StreamlineLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # RenderingCfg
    def StreamlineSpeed(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # RenderingCfg
    def Target(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            x = o + self._tab.Pos
            from SEAScope.API.RenderableId import RenderableId
            obj = RenderableId()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # RenderingCfg
    def FilterMode(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # RenderingCfg
    def BillboardsSize(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # RenderingCfg
    def BillboardsDensity(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(36))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # RenderingCfg
    def LineThickness(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(38))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

def RenderingCfgStart(builder): builder.StartObject(18)
def RenderingCfgAddRendered(builder, rendered): builder.PrependBoolSlot(0, rendered, 0)
def RenderingCfgAddLogscale(builder, logscale): builder.PrependBoolSlot(1, logscale, 0)
def RenderingCfgAddMin(builder, min): builder.PrependFloat32Slot(2, min, 0.0)
def RenderingCfgAddMax(builder, max): builder.PrependFloat32Slot(3, max, 0.0)
def RenderingCfgAddOpacity(builder, opacity): builder.PrependFloat32Slot(4, opacity, 0.0)
def RenderingCfgAddZindex(builder, zindex): builder.PrependFloat32Slot(5, zindex, 0.0)
def RenderingCfgAddColor(builder, color): builder.PrependStructSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(color), 0)
def RenderingCfgAddColormap(builder, colormap): builder.PrependUOffsetTRelativeSlot(7, flatbuffers.number_types.UOffsetTFlags.py_type(colormap), 0)
def RenderingCfgAddRenderMethod(builder, renderMethod): builder.PrependInt8Slot(8, renderMethod, 0)
def RenderingCfgAddParticlesCount(builder, particlesCount): builder.PrependUint32Slot(9, particlesCount, 0)
def RenderingCfgAddParticleTTL(builder, particleTTL): builder.PrependUint32Slot(10, particleTTL, 0)
def RenderingCfgAddStreamlineLength(builder, streamlineLength): builder.PrependUint32Slot(11, streamlineLength, 0)
def RenderingCfgAddStreamlineSpeed(builder, streamlineSpeed): builder.PrependFloat32Slot(12, streamlineSpeed, 0.0)
def RenderingCfgAddTarget(builder, target): builder.PrependStructSlot(13, flatbuffers.number_types.UOffsetTFlags.py_type(target), 0)
def RenderingCfgAddFilterMode(builder, filterMode): builder.PrependInt8Slot(14, filterMode, 0)
def RenderingCfgAddBillboardsSize(builder, billboardsSize): builder.PrependFloat32Slot(15, billboardsSize, 0.0)
def RenderingCfgAddBillboardsDensity(builder, billboardsDensity): builder.PrependFloat32Slot(16, billboardsDensity, 0.0)
def RenderingCfgAddLineThickness(builder, lineThickness): builder.PrependFloat32Slot(17, lineThickness, 0.0)
def RenderingCfgEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: API

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class RemoveCollection(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsRemoveCollection(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = RemoveCollection()
        x.Init(buf, n + offset)
        return x

    # RemoveCollection
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # RemoveCollection
    def CollectionId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = o + self._tab.Pos
            from SEAScope.API.RenderableId import RenderableId
            obj = RenderableId()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

def RemoveCollectionStart(builder): builder.StartObject(1)
def RemoveCollectionAddCollectionId(builder, collectionId): builder.PrependStructSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(collectionId), 0)
def RemoveCollectionEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: API

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Variable(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsVariable(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Variable()
        x.Init(buf, n + offset)
        return x

    # Variable
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Variable
    def Id(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # Variable
    def Label(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Variable
    def Units(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Variable
    def Fields(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # Variable
    def FieldsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Variable
    def FieldsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        return o == 0

    # Variable
    def Tags(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from SEAScope.API.Tag import Tag
            obj = Tag()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Variable
    def TagsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Variable
    def TagsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0

    # Variable
    def DefaultRenderingMethod(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

def VariableStart(builder): builder.StartObject(6)
def VariableAddId(builder, id): builder.PrependUint32Slot(0, id, 0)
def VariableAddLabel(builder, label): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(label), 0)
def VariableAddUnits(builder, units): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(units), 0)
def VariableAddFields(builder, fields): builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(fields), 0)
def VariableStartFieldsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def VariableAddTags(builder, tags): builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(tags), 0)
def VariableStartTagsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def VariableAddDefaultRenderingMethod(builder, defaultRenderingMethod): builder.PrependInt8Slot(5, defaultRenderingMethod, 0)
def VariableEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: API

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Color(object):
    __slots__ = ['_tab']

    # Color
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Color
    def R(self): return self._tab.Get(flatbuffers.number_types.Uint32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(0))
    # Color
    def G(self): return self._tab.Get(flatbuffers.number_types.Uint32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(4))
    # Color
    def B(self): return self._tab.Get(flatbuffers.number_types.Uint32Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(8))

def CreateColor(builder, r, g, b):
    builder.Prep(4, 12)
    builder.PrependUint32(b)
    builder.PrependUint32(g)
    builder.PrependUint32(r)
    return builder.Offset()



# vim: ts=4:sts=4:sw=4
#
# @author: <sylvain.herledan@oceandatalab.com>
# @date: 2016-09-09
#
# This file is part of SEAScope, a 3D visualisation and analysis application
# for satellite, in-situ and numerical model data.
#
# Copyright (C) 2014-2023 OceanDataLab
#
# SEAScope is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SEAScope. If not, see <https://www.gnu.org/licenses/>.

"""
This module handles the serialization and deserialization of rendering
configuration objects
"""

import logging
import flatbuffers
import SEAScope.API.RenderingCfg
import SEAScope.types.color
import SEAScope.types.renderable_id
import SEAScope.API.FilterMode

logger = logging.getLogger(__name__)

rendering_methods = {
    'RASTER': SEAScope.API.RenderingMethod.RenderingMethod.RASTER,
    'ARROWS': SEAScope.API.RenderingMethod.RenderingMethod.ARROWS,
    'BARBS': SEAScope.API.RenderingMethod.RenderingMethod.BARBS,
    'TRAJECTORIES': SEAScope.API.RenderingMethod.RenderingMethod.TRAJECTORIES,
    'DOTSCLOUD': SEAScope.API.RenderingMethod.RenderingMethod.DOTSCLOUD,
    'STREAMLINES': SEAScope.API.RenderingMethod.RenderingMethod.STREAMLINES,
    'RAWRGB': SEAScope.API.RenderingMethod.RenderingMethod.RAWRGB}
"""dict: Methods for rendering data in SEAScope.

Note that DOTSCLOUD is not supported by the current version of SEAScope
"""

filter_modes = {
    'NEAREST': SEAScope.API.FilterMode.FilterMode.NEAREST,
    'BILINEAR': SEAScope.API.FilterMode.FilterMode.BILINEAR}
"""dict: Rendering filters available in SEAScope. These filters are applied for
mapping pixels with data."""


def serialize(builder, cfg_obj):
    """
    Serialize a rendering configuration using FlatBuffers.

    Parameters
    ----------
    builder : flatbuffers.builder.Builder
        The FlatBuffers builder instance which serializes data. If this
        parameter is None, then a new builder will be created
    cfg_obj : dict
        Dictionary which contains information about the rendering configuration
        to serialize

        It must have 15 keys:

        - ``rendered`` : a :obj:`bool` telling SEAScope if the associated
          target should be rendered or not (only relevant for collections)

        - ``logscale`` : a :obj:`bool` telling SEAScope to use a logarithmic
          scale when applying the colormap (True) or the default linear scale
          (False)

        - ``min`` : a :obj:`float` that SEAScope must use as the minimal value
          when applying the colormap

        - ``max`` : a :obj:`float` that SEAScope must use as the maximal value
          when applying the colormap

        - ``opacity`` : a :obj:`float` giving the opacity of the target. Its
          value belongs to the [0.0, 1.0] range: 0.0 is fully transparent, 1.0
          is completely opaque

        - ``zindex`` : a :obj:`float` that SEAScope uses to decide in which
          order the data should be rendered. Granules with a low zindex are
          rendered before granules with a high zindex

        - ``colormap`` : a :obj:`str` which identifies the colormap applied to
          the representation of the variable.

          A colormap identifier is only valid if there is a file in the
          ``colormaps`` directory whose name is the identifier suffixed by the
          ``.rgb`` extension.

        - ``renderMethod`` : a :obj:`str` that tells SEAScope which rendering
          method must be used to display the variable. The value should be a
          key of the :const:`SEAScope.types.rendering_cfg.rendering_methods`
          dictionary

        - ``particlesCount`` : an :obj:`int` which sets the size of the pool of
          animated particles.

          This value controls the density of the particles when using the
          streamlines rendering method.

          Increasing the density adds a toll on the GPU and may significatively
          reduce fluidity.

        - ``particleTTL`` : an :obj:`int` (deprecated)

        - ``streamlineLength`` : an :obj:`int` which tells SEAScope how many
          segments each streamline should be made of

        - ``streamlineSpeed`` : a :obj:`float` (deprecated)

        - ``filterMode`` : a :obj:`str` that tells SEAScope which filter to use
          when mapping pixels with data values. The value must be a key of the
          :const:`SEAScope.types.rendering_cfg.filter_modes` dictionary

        - ``target`` : a :obj:`dict` that identifies the target of the
          rendering configuration. The value must satisfy the requirements of
          the :func:`SEAScope.types.renderable_id.serialize` method.

        - ``color`` : a :obj:`dict` describing the uniform color that SEAScope
          will apply when rendering the target. The value must satisfy the
          requirements of the :func:`SEAScope.types.color.serialize` method.

          If both ``color`` and ``colormap`` are defined, then the colormap
          will be used and the color will be ignored.

        - ``billboardsSize`` : a :obj:float used by SEAScope to define the size
          of symbols when rendering data as arrows or barbs.

        - ``billboardsDensity`` : a :obj:float value between 0.0 and 1.0 that
          SEAScope uses to compute how many symbols it must draw when using the
          arrows or barbs renderers. A value of 0.0 means that symbols will be
          separated by a blank space which has the same size as one symbol.
          A value of 1.0 means that there is no blank space between visible
          symbols.

        - ``lineThickness`` : a :obj:float that tells SEAScope how thick lines
          should be (in pixels) when drawing trajectories or polylines.

    Returns
    -------
    tuple(flatbuffers.builder.Builder, int)
        A tuple which contains two elements:

        - the :obj:`flatbuffers.builder.Builder` instance which has been used
          to serialize data
        - an :obj:`int` which is the address/offset of the serialized object
          in the builder buffer
    """
    if builder is None:
        builder = flatbuffers.Builder(0)

    cls = SEAScope.API.RenderingCfg

    _rendered = cfg_obj['rendered']
    _logscale = cfg_obj['logscale']
    _min = cfg_obj['min']
    _max = cfg_obj['max']
    _opacity = cfg_obj['opacity']
    _zindex = cfg_obj['zindex']
    _colormap = builder.CreateString(cfg_obj['colormap'])
    _render_method = rendering_methods.get(cfg_obj['renderMethod'], None)
    _particles_count = int(cfg_obj['particlesCount'])
    _particle_ttl = int(cfg_obj['particleTTL'])
    _streamline_length = int(cfg_obj['streamlineLength'])
    _streamline_speed = cfg_obj['streamlineSpeed']
    _filter_mode = filter_modes.get(cfg_obj['filterMode'], None)
    _billboards_size = cfg_obj['billboardsSize']
    _billboards_density = cfg_obj['billboardsDensity']
    _line_thickness = cfg_obj['lineThickness']

    cls.RenderingCfgStart(builder)
    cls.RenderingCfgAddRendered(builder, _rendered)
    cls.RenderingCfgAddLogscale(builder, _logscale)
    cls.RenderingCfgAddMin(builder, _min)
    cls.RenderingCfgAddMax(builder, _max)
    cls.RenderingCfgAddOpacity(builder, _opacity)
    cls.RenderingCfgAddZindex(builder, _zindex)
    cls.RenderingCfgAddColormap(builder, _colormap)
    cls.RenderingCfgAddRenderMethod(builder, _render_method)
    cls.RenderingCfgAddParticlesCount(builder, _particles_count)
    cls.RenderingCfgAddParticleTTL(builder, _particle_ttl)
    cls.RenderingCfgAddStreamlineLength(builder, _streamline_length)
    cls.RenderingCfgAddStreamlineSpeed(builder, _streamline_speed)
    cls.RenderingCfgAddFilterMode(builder, _filter_mode)
    cls.RenderingCfgAddBillboardsSize(builder, _billboards_size)
    cls.RenderingCfgAddBillboardsDensity(builder, _billboards_density)
    cls.RenderingCfgAddLineThickness(builder, _line_thickness)

    _, _target = SEAScope.types.renderable_id.serialize(builder,
                                                        cfg_obj['target'])
    cls.RenderingCfgAddTarget(builder, _target)
    _, _color = SEAScope.types.color.serialize(builder, cfg_obj['color'])
    cls.RenderingCfgAddColor(builder, _color)
    cfg = cls.RenderingCfgEnd(builder)
    return builder, cfg


def deserialize(o):
    """
    Rebuild a rendering configuration from a FlatBuffers buffer.

    Parameters
    ----------
    buf : bytearray
        The buffer which contain the rendering configuration object serialized
        with FlatBuffers

    Returns
    -------
    dict
        The deserialized rendering configuration object as a dictionary.
    """
    cfg = {}
    cfg['rendered'] = 0 < o.Rendered()
    cfg['logscale'] = 0 < o.Logscale()
    cfg['min'] = o.Min()
    cfg['max'] = o.Max()
    cfg['opacity'] = o.Opacity()
    cfg['zindex'] = o.Zindex()
    cfg['color'] = SEAScope.types.color.deserialize(o.Color())
    cfg['colormap'] = o.Colormap()
    rendering = o.RenderMethod()
    for k, v in rendering_methods.items():
        if v == rendering:
            cfg['renderMethod'] = k
            break
    filter_mode = o.FilterMode()
    for k, v in filter_modes.items():
        if v == filter_mode:
            cfg['filterMode'] = k
            break
    cfg['particlesCount'] = o.ParticlesCount()
    cfg['particleTTL'] = o.ParticleTTL()
    cfg['streamlineLength'] = o.StreamlineLength()
    cfg['streamlineSpeed'] = o.StreamlineSpeed()
    cfg['billboardsSize'] = o.BillboardsSize()
    cfg['billboardsDensity'] = o.BillboardsDensity()
    cfg['lineThickness'] = o.LineThickness()
    cfg['target'] = SEAScope.types.renderable_id.deserialize(o.Target())

    return cfg



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: API

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Shape(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsShape(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Shape()
        x.Init(buf, n + offset)
        return x

    # Shape
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Shape
    def XArity(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # Shape
    def YArity(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # Shape
    def Gcps(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 24
            from SEAScope.API.GCP import GCP
            obj = GCP()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Shape
    def GcpsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Shape
    def GcpsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

def ShapeStart(builder): builder.StartObject(3)
def ShapeAddXArity(builder, xArity): builder.PrependUint32Slot(0, xArity, 0)
def ShapeAddYArity(builder, yArity): builder.PrependUint32Slot(1, yArity, 0)
def ShapeAddGcps(builder, gcps): builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(gcps), 0)
def ShapeStartGcpsVector(builder, numElems): return builder.StartVector(24, numElems, 8)
def ShapeEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: API

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class GetExtractionResponse(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsGetExtractionResponse(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = GetExtractionResponse()
        x.Init(buf, n + offset)
        return x

    # GetExtractionResponse
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # GetExtractionResponse
    def Extractions(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from SEAScope.API.Extraction import Extraction
            obj = Extraction()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # GetExtractionResponse
    def ExtractionsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # GetExtractionResponse
    def ExtractionsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

def GetExtractionResponseStart(builder): builder.StartObject(1)
def GetExtractionResponseAddExtractions(builder, extractions): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(extractions), 0)
def GetExtractionResponseStartExtractionsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def GetExtractionResponseEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: API

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class SelectVariable(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsSelectVariable(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = SelectVariable()
        x.Init(buf, n + offset)
        return x

    # SelectVariable
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # SelectVariable
    def Id(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = o + self._tab.Pos
            from SEAScope.API.RenderableId import RenderableId
            obj = RenderableId()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # SelectVariable
    def Selected(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # SelectVariable
    def Exclusive(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

def SelectVariableStart(builder): builder.StartObject(3)
def SelectVariableAddId(builder, id): builder.PrependStructSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(id), 0)
def SelectVariableAddSelected(builder, selected): builder.PrependBoolSlot(1, selected, 0)
def SelectVariableAddExclusive(builder, exclusive): builder.PrependBoolSlot(2, exclusive, 0)
def SelectVariableEnd(builder): return builder.EndObject()



# -*- coding: utf-8 -*-

import os
import sys
import logging
from setuptools import setup
from setuptools.command.test import test as TestCommand

# Setup logging
logger = logging.getLogger()
handler = logging.StreamHandler()
handler.setLevel(logging.WARNING)
logger.addHandler(handler)
logger.setLevel(logging.WARNING)

__package_name__ = 'SEAScope'
project_dir = os.path.dirname(__file__)

# Use whichcraft to make sure that the "which" command is available
# https://github.com/pydanny/whichcraft/blob/master/whichcraft.py
# BSD-3 clauses licence
sys.path.insert(0, os.path.abspath(project_dir))
from whichcraft import which  # noqa


class Tox(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_siote = True

    def run_tests(self):
        import tox
        errcode = tox.cmdline(self.test_args)
        sys.exit(errcode)


git_exe = which('git')
git_dir = os.path.join(project_dir, '.git')
has_git = (git_exe is not None and os.path.isdir(git_dir))
version_path = os.path.join(project_dir, 'VERSION.txt')
readme_path = os.path.join(project_dir, 'README_short.rst')
package_dir = os.path.join(project_dir, __package_name__)
init_path = os.path.join(package_dir, '__init__.py')

# Read metadata from the main __init__.py file
metadata = {}
with open(init_path, 'rt') as f:
    exec(f.read(), metadata)

# Regenerate a version file from git history
if has_git:
    try:
        import subprocess32 as subprocess
    except ImportError:
        import subprocess

    gitrev = (git_exe, 'rev-list', 'HEAD', '--count')
    major, minor = metadata['__version__'].split('.')
    commits = subprocess.check_output(gitrev).decode('utf-8').strip()
    with open(version_path, 'wt') as f:
        f.write('{}.{}.{}'.format(major, minor, commits))

# Refuse to install package if version is not available
if not os.path.isfile(version_path):
    logger.error('Version file {} missing.'.format(version_path))
    logger.error('Please use a proper release of the code.')
    sys.exit(1)

with open(version_path, 'rt') as f:
    version = f.read()

with open(readme_path, 'rt') as f:
    long_description = f.read()

processor_deps = ('matplotlib', 'numpy', 'pyproj')
doc_deps = ('sphinx', 'sphinxcontrib-napoleon')

setup(
    zip_safe=False,
    name='pySEAScope',
    version=version,
    author=metadata['__author__'],
    author_email=metadata['__author_email__'],
    maintainer=metadata['__author__'],
    maintainer_email=metadata['__author_email__'],
    url=metadata['__url__'],
    packages=(__package_name__,
              '{}.API'.format(__package_name__),
              '{}.cmds'.format(__package_name__),
              '{}.types'.format(__package_name__),
              '{}.lib'.format(__package_name__),
              '{}.cli'.format(__package_name__)),
    scripts=[],
    license='COPYING.LESSER',
    description=metadata['__description__'],
    long_description=long_description,
    long_description_content_type='text/x-rst',
    install_requires=['flatbuffers<2.0'],
    tests_require=['tox', 'virtualenv'],
    extras_require={'processor': processor_deps, 'doc': doc_deps},
    classifiers=metadata['__classifiers__'],
    cmdclass={'test': Tox},
    entry_points={'console_scripts': (
        'seascope-processor = SEAScope.cli.server:seascope_processor')}
)



# vim: ts=4:sts=4:sw=4
#
# @author: <sylvain.herledan@oceandatalab.com>
# @date: 2017-05-02
#
# This file is part of SEAScope, a 3D visualisation and analysis application
# for satellite, in-situ and numerical model data.
#
# Copyright (C) 2014-2023 OceanDataLab
#
# SEAScope is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
"""
This module contains the implementation of the processing server that can be
started from the command line with ``seascope-processor``
"""

import os
import sys
import numpy
import struct
import socket
import logging
import datetime
import tempfile
import threading

import SEAScope.lib
import SEAScope.lib.utils
import SEAScope.lib.plots

# Initialize Matplotlib backend
import matplotlib
matplotlib.use('Agg')


logger = logging.getLogger(__name__)


def serialize_extraction():
    """Handler called to fetch extracted data from SEAScope and save them in a
    file using the Numpy serialization format.

    Returns
    -------
    tuple(bool, str)
        A tuple which contains:

        - a :obj:`bool` set to True if the operation succeded
        - a :obj:`str` message containing the path of the output file
    """
    logger.info('Serializing extracted data...')
    now = datetime.datetime.utcnow()
    output_dir = tempfile.mkdtemp(suffix=now.strftime('%Y%m%dT%H%M%S'))
    output_path = os.path.join(output_dir, 'seascope_extract.pyo')

    granules = SEAScope.lib.get_extracted_data()
    SEAScope.lib.save_pyo(granules, output_path)

    logger.info('Done.')
    return True, '=> Serialization output saved in {}'.format(output_path)


def plot_extraction():
    """Handler called to fetch extracted data from SEAScope, render them as
    2D plots and save the result as PNG images.

    Returns
    -------
    tuple(bool, str)
        A tuple which contains:

        - a :obj:`bool` set to True if the operation succeded
        - a :obj:`str` message containing the path of the output directory
          where the PNG files have been saved
    """
    logger.info('Saving extracted data as PNGs...')
    now = datetime.datetime.utcnow()
    output_dir = tempfile.mkdtemp(suffix=now.strftime('%Y%m%dT%H%M%S'))

    granules = SEAScope.lib.get_extracted_data()

    for granule_uri in granules:
        granule_data = granules[granule_uri]['data']
        for field in granule_data:
            _data = numpy.flipud(granule_data[field])
            _uri = os.path.basename(granule_uri)

            output_file = '{}_{}.png'.format(_uri, field)
            output_path = os.path.join(output_dir, output_file)

            SEAScope.lib.plots.plot_2d_data(_uri, field, _data, output_path)

    return True, '=> PNGs generated for all granules in {}'.format(output_dir)


def plot_transect():
    """Handler called to fetch extracted data from SEAScope, render them as
    transects in a single plot and save the result as a PNG image.

    Returns
    -------
    tuple(bool, str)
        A tuple which contains:

        - a :obj:`bool` set to True if the operation succeded
        - a :obj:`str` message containing the path of the output PNG file
    """
    logger.info('Saving extracted data as PNGs...')
    now = datetime.datetime.utcnow()
    output_dir = tempfile.mkdtemp(suffix=now.strftime('%Y%m%dT%H%M%S'))

    granules = SEAScope.lib.get_extracted_data()

    if 'mask' not in granules:
        raise Exception('Missing mask in transects')

    dists = SEAScope.lib.get_dists(granules['mask'])

    transects = {}
    for granule_uri in granules:
        if 'mask' == granule_uri:
            continue

        granule = granules[granule_uri]
        for field in granule['data']:
            _uri = os.path.basename(granule_uri)
            uri = '{}: {}'.format(_uri, field)
            transects[uri] = granule['data'][field][0, :]

    output_file = '{}.png'.format('transects')
    output_path = os.path.join(output_dir, output_file)

    SEAScope.lib.plots.plot_transect(dists, transects, output_path)

    return True, '=> Result saved in {}'.format(output_path)


class ProcessingUnit(threading.Thread):
    """Thread that handles a single request sent by SEAScope.

    Parameters
    ----------
    s : socket:socket
        The socket connected to SEAScope
    """

    def __init__(self, s):
        super(ProcessingUnit, self).__init__()
        self.socket = s

    def run(self):
        """
        Method called when the thread execution starts.

        It parses the request sent by SEAScope, extracts an opcode and
        calls a processing handler accordingly.
        """
        buf = self.socket.recv(8)
        _opcode = struct.unpack('<Q', buf)
        opcode = _opcode[0]
        logger.debug('Received opcode {}'.format(opcode))
        if 0 == opcode:
            # noop
            pass
        elif 1 == opcode:
            ok, msg = plot_extraction()
            if not ok:
                logger.error(msg)
                return
            logger.info(msg)
        elif 2 == opcode:
            ok, msg = plot_transect()
            if not ok:
                logger.error(msg)
                return
            logger.info(msg)
        elif 3 == opcode:
            ok, msg = serialize_extraction()
            if not ok:
                logger.error(msg)
                return
            logger.info(msg)
        else:
            logger.error('Received unknown opcode {}'.format(opcode))

    def close(self):
        """Method called on thread termination.

        It makes sure that the socket is properly closed before the object is
        garbage-collected.
        """
        self.socket.close()


def bind_socket(s, interface='127.0.0.1'):
    """Bind a socket to an interface.

    The method tries to bind the default port (53450) first but it will
    fallback to the first port available if 53450 is already in use.

    Parameters
    ----------
    s : socket.socket
        The socket to bind
    interface : string
        The IP address to which the socket must be bound to


    Returns
    -------
    tuple(str, int)
        A tuple which contains two elements:

        - a :obj:`str` for the address to which the socket is actually bound to
        - a :obj:`int` for the port that the socket has managed to acquire
    """
    # Check default port first
    use_default = True
    port = 53450
    try:
        s.bind((interface, port))
        logger.info('Using default port {}'.format(port))
    except socket.error:
        # Try to get a free port
        logger.info('Default port {} already in use'.format(port))
        logger.info('Requesting another one...')
        use_default = False
        port = 0
        s.bind((interface, port))

    addr, port = s.getsockname()
    if use_default is False:
        logger.warn('\n{}'.format(''.join(['-'] * 80)))
        logger.warn('* seascope-processor is not listening on the default '
                    'port (already in use).\n*\n* You will have to '
                    'update SEAScope configuration manually for it to send \n'
                    '* processing requests to the right port: {}'.format(port))
        logger.warn('{}\n'.format(''.join(['-'] * 80)))
    return (addr, port)


def serve_forever():
    """Starts a sockets server that handles requests sent by SEAScope and
    creates a :class:`SEAScope.cli.server.ProcessingUnit` for each request to
    process.

    Returns
    -------
    tuple(bool, str)
        A tuple which contains two elements:

        - A :obj:`bool` set to False if an error occurs, True if everything
          went fine
        - A obj:`str` that contains the error message, or None if not relevant
    """
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        addr, port = bind_socket(s)
        s.listen(15)
    except socket.error:
        _, e, _ = sys.exc_info()
        logger.error('{}'.format(e))
        return False, 'Failed to create socket'

    # TODO: find a way to tell SEAScope about the port
    logger.info('Listening on {}:{}'.format(addr, port))
    logger.info('Processing server is running.')
    while True:
        try:
            client_socket, address = s.accept()
            processing_unit = ProcessingUnit(client_socket)
            processing_unit.start()
        except KeyboardInterrupt:
            return True, None

    return False, 'Something went wrong'


def seascope_processor():
    """Method called when the user executes ``seascope-processor`` on the
    command line.

    It setups logging before starting the sockets server.
    """
    # Setup logging
    main_logger = logging.getLogger()
    main_logger.handlers = []
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    main_logger.addHandler(handler)
    main_logger.setLevel(logging.INFO)

    ok, msg = serve_forever()
    if not ok:
        logger.error(msg)
        sys.exit(1)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: API

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class DataInfo(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsDataInfo(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = DataInfo()
        x.Init(buf, n + offset)
        return x

    # DataInfo
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # DataInfo
    def Channels(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # DataInfo
    def XArity(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # DataInfo
    def YArity(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # DataInfo
    def DataType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # DataInfo
    def Offsets(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Float32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # DataInfo
    def OffsetsAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Float32Flags, o)
        return 0

    # DataInfo
    def OffsetsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # DataInfo
    def OffsetsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0

    # DataInfo
    def ScaleFactors(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Float32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # DataInfo
    def ScaleFactorsAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Float32Flags, o)
        return 0

    # DataInfo
    def ScaleFactorsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # DataInfo
    def ScaleFactorsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        return o == 0

    # DataInfo
    def FillValues(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Float32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # DataInfo
    def FillValuesAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Float32Flags, o)
        return 0

    # DataInfo
    def FillValuesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # DataInfo
    def FillValuesIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        return o == 0

def DataInfoStart(builder): builder.StartObject(7)
def DataInfoAddChannels(builder, channels): builder.PrependUint32Slot(0, channels, 0)
def DataInfoAddXArity(builder, xArity): builder.PrependUint32Slot(1, xArity, 0)
def DataInfoAddYArity(builder, yArity): builder.PrependUint32Slot(2, yArity, 0)
def DataInfoAddDataType(builder, dataType): builder.PrependInt8Slot(3, dataType, 0)
def DataInfoAddOffsets(builder, offsets): builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(offsets), 0)
def DataInfoStartOffsetsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def DataInfoAddScaleFactors(builder, scaleFactors): builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(scaleFactors), 0)
def DataInfoStartScaleFactorsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def DataInfoAddFillValues(builder, fillValues): builder.PrependUOffsetTRelativeSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(fillValues), 0)
def DataInfoStartFillValuesVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def DataInfoEnd(builder): return builder.EndObject()



# vim: set ts=4:sts=4:sw=4
#
# @author: <sylvain.herledan@oceandatalab.com>
#
# This file is part of SEAScope, a 3D visualisation and analysis application
# for satellite, in-situ and numerical model data.
#
# Copyright (C) 2014-2023 OceanDataLab
#
# SEAScope is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SEAScope. If not, see <https://www.gnu.org/licenses/>.

"""
This module provides methods for creating and reformatting data structures.
"""
import math
import logging
import datetime

logger = logging.getLogger(__name__)


class DynamicPackingFloatIncompatbilityError(Exception):
    """Raised when :func:`SEAScope.lib.utils.set_field` is called with both
    ``dynamic_packing`` and ``as_float`` set to True`: dynamic packing is only
    implemented for unsigned bytes"""
    pass


class InvalidShapeError(ValueError):
    """Raised when the shape passed to :func:`SEAScope.lib.utils.get_lonlat`
    does not have exactly 2 elements (shape for a 2D matrix)."""
    pass


class FieldShapeError(ValueError):
    """Raised when the data associated with a field is neither 1D nor 2D."""
    pass


class IncompatibleFieldError(ValueError):
    """Raised when a field cannot be added to a granule because it is not
    compatible with already attached fields (shape mismatch)."""
    pass


def _id_generator(offset):
    """"""
    for value in range(offset, 100000):
        yield value


collection_id_generator = _id_generator(10)
granule_id_generator = _id_generator(1000)


def init_ids(collection_offset, granule_offset):
    """Initialize the local, i.e. specific to current data source, identifiers
    generators for collections and granules.

    Parameters
    ----------
    collection_offset : int
        Initial value for the collection identifiers generator. Please note
        that the initial value cannot be less than 3 because identifiers "1"
        and "2" are reserved for the "User polygons" and "User polylines"
        collections that SEAScope creates automatically (these two collections
        share the same source as the ones you create using the Python bindings)
    granule_offset : int
        Initial value for the granule identifiers generator. Please note that
        the initial value cannot be less than 1000 so that SEAScope has enough
        reserved identifiers for the polygons and polylines drawn by the user.
    """
    global collection_id_generator
    global granule_id_generator

    # collection IDs 1 and 2 are reserved for polygons and polylines
    if collection_offset < 3:
        collection_offset = 3
        logger.warning('Collection IDs 1 and 2 are reserved for internal use '
                       'in SEAScope. Your input has been modified to prevent '
                       'issues: collection_offset = 3')

    # Reserve 1000 granule IDs for the polygons and polylines the user will
    # draw in SEAScope
    if granule_offset < 1000:
        granule_offset = 1000
        logger.warning('Granule IDs below 1000 are reserved for internal use '
                       'in SEAScope. Your input has been modified to prevent '
                       'issues: granule_offset = 1000')

    collection_id_generator = _id_generator(collection_offset)
    granule_id_generator = _id_generator(granule_offset)


def create_collection(label):
    """Helper method to create the dictionary that represents a collection.

    Most fields are initialized with default values that you can/should
    customize before passing the generated dictionary to other methods.

    Parameters
    ----------
    label : str
        Label describing the collection. This is the text that SEAScope will
        display in the catalogue for this collection.

    Returns
    -------
    tuple(int, dict)
        A tuple which contains two elements:

          - an integer which is the local identifier, i.e. the identifier
            within a source, for the collection

          - a dictionary which contains all the information SEAScope needs to
            identify and format data associated with the collection variables.

    Example
    -------
    >>> import SEAScope.lib.utils
    >>> result = SEAScope.lib.utils.create_collection('A custom collection')
    >>> coll_id, coll_dict = result
    >>> print(coll_id)
    10
    >>> print(coll_dict)
    {'id': {'granuleLevel': False,
            'sourceId': 1,
            'collectionId': 10,
            'granuleId': 0,
            'variableId': 0},
     'mustBeCurrent': False,
     'xSeamless': False,
     'ySeamless': False,
     'NEWSAligned': False,
     'label': 'A custom collection',
     'tags': {},
     'variables': [],
     'variablesList': {},
     'defaultVariable': 0}
    """
    unique_id = next(collection_id_generator)
    _rid = {'granuleLevel': False,
            'sourceId': 1,
            'collectionId': unique_id,
            'granuleId': 0,
            'variableId': 0}
    result = {'id': _rid,
              'mustBeCurrent': False,
              'xSeamless': False,
              'ySeamless': False,
              'NEWSAligned': False,
              'label': label,
              'tags': {},
              'variables': [],
              'variablesList': {},
              'defaultVariable': 0}
    return unique_id, result


def create_variable(collection, label, fields, units='', dims=2):
    """Helper method to create the dictionary that represents a variable.

    Most fields are initialized with default values that you can/should
    customize before passing the generated dictionary to other methods.

    Parameters
    ----------
    collection : dict
        Dictionary which represents a collection. This dictionary is either
        handcrafted or generated by the :func:`create_collection` method.
    label : str
        Label describing the variable. This is the text that will be displayed
        in the SEAScope catalogue.
    units : str
        Units of the geophysical data contained in the variable. Defaults to an
        empty string.
    dims : int
        Number of dimensions of the data matrix associated with the variable.
        If this parameter is set to 1 the default rendering method for this
        variable will be TRAJECTORIES, otherwise it will be RASTER.

    Returns
    -------
    dict
        A dictionary which contains all the information SEAScope needs to
        identify the variable and render the associated data.

    Example
    -------
    >>> import SEAScope.lib.utils
    >>> result = SEAScope.lib.utils.create_collection('A custom collection')
    >>> coll_id, coll_dict = result
    >>> var_dict = SEAScope.lib.utils.create_variable(coll_dict, 'my variable',
    >>>                                               ['u_field', 'v_field'],
    >>>                                               'm.s-1')
    >>> print(var_dict)
    {'id': 0,
     'label': 'my variable',
     'units': 'm.s-1',
     'fields': ['u_field', 'v_field'],
     'defaultRenderingMethod': 'RASTER',
     'tags': {},
     'collection': {'granuleLevel': False,
                    'sourceId': 1,
                    'collectionId': 10,
                    'granuleId': 0,
                    'variableId': 0},
     'rendering': {'rendered': True,
                   'logscale': False,
                   'min': 0,
                   'max': 1,
                   'opacity': 1.0,
                   'zindex': 500,
                   'color': [50, 12, 87],
                   'colormap': 'grayscale',
                   'renderMethod': 'RASTER',
                   'particlesCount': 1000,
                   'particleTTL': 10,
                   'streamlineLength': 20,
                   'streamlineSpeed': 0.0,
                   'filterMode': 'NEAREST',
                   'target': {'granuleLevel': False,
                              'sourceId': 1,
                              'collectionId': 10,
                              'granuleId': 0,
                              'variableId': 0}}}
    """
    if label in collection['variablesList']:
        var_id = collection['variablesList'][label]
        logger.warn('Variable "{}" already exists for this collection and has '
                    'identifier "{}"'.format(label, var_id))
        return None

    unique_id = len(collection['variablesList'].keys())
    collection['variablesList'][label] = unique_id

    rcfg = create_rendering_config(collection['id']['collectionId'], unique_id)
    r_method = 'RASTER'
    if 1 == dims:
        r_method = 'TRAJECTORIES'
    rcfg['renderMethod'] = r_method

    result = {'id': unique_id,
              'label': label,
              'units': units,
              'fields': fields,
              'defaultRenderingMethod': r_method,
              'tags': {},
              'collection': collection['id'],
              'rendering': rcfg}
    return result


def create_rendering_config(col_id=0, var_id=0):
    """Helper method to create the dictionary that represents a rendering
    configuration (for a variable).

    Most fields are initialized with default values that you can/should
    customize before passing the generated dictionary to other methods.

    Parameters
    ----------
    col_id : int
        Local (i.e. specifc to the current data source) identifier for the
        collection
    var_id : int
        Identifier for the variable within the collection

    Returns
    -------
    dict
        A dictionary which contains the rendering parameters to use for the
        variable which corresponds to the (col_id, var_id) couple.

    Example
    -------
    >>> import SEAScope.lib.utils
    >>> rcfg = SEAScope.lib.utils.create_rendering_config(533, 5)
    >>> print(rcfg)
    {'rendered': True,
     'logscale': False,
     'min': 0,
     'max': 1,
     'opacity': 1.0,
     'zindex': 500,
     'color': [50, 12, 87],
     'colormap': 'grayscale',
     'renderMethod': 'RASTER',
     'particlesCount': 1000,
     'particleTTL': 10,
     'streamlineLength': 20,
     'streamlineSpeed': 0.0,
     'filterMode': 'NEAREST',
     'billboardsSize': 32,
     'billboardsDensity': 0,
     'lineThickness': 1,
     'target': {'granuleLevel': False,
                'sourceId': 1,
                'collectionId': 533,
                'granuleId': 0,
                'variableId': 5}}
    """
    target = {'granuleLevel': False,
              'sourceId': 1,
              'collectionId': col_id,
              'granuleId': 0,
              'variableId': var_id}

    result = {'rendered': True,
              'logscale': False,
              'min': 0,
              'max': 1,
              'opacity': 1.0,
              'zindex': 500,
              'color': [50, 12, 87],
              'colormap': 'grayscale',
              'renderMethod': 'RASTER',
              'particlesCount': 1000,
              'particleTTL': 10,
              'streamlineLength': 20,
              'streamlineSpeed': 0.0,
              'filterMode': 'NEAREST',
              'billboardsSize': 32,
              'billboardsDensity': 0,
              'lineThickness': 1,
              'target': target}
    return result


def create_granule(collection_id, gcps, start, stop):
    """Helper method to create the dictionary that represents a granule.

    Most fields are initialized with default values that you can/should
    customize before passing the generated dictionary to other methods.

    Time coverage is [start, stop[

    Parameters
    ----------
    collection_id : int
        Local (i.e. specifc to the current data source) identifier for the
        collection of the granule
    gcps : list of dict
        List of Ground Control Points, represented as dictionaries with four
        keys: lon, lat, i and j, that give the spatial coverage of the
        granule and provide intermediary points to map data on a sphere
    start : datetime.datetime
        Begining of the time coverage of the granule, in UTC (included)
    stop : datetime.datetime
        End of the time coverage of the granule, in UTC (excluded)

    Returns
    -------
    tuple(int, dict)
        A tuple which contains two elements:

          - an integer which is the local identifier, i.e. the identifier
            within a source, for the granule

          - a dictionary which contains all the information SEAScope needs to
            identify and format data associated with the granule.
    """
    granule_id = next(granule_id_generator)
    _start = (start - datetime.datetime(1970, 1, 1)).total_seconds() * 1000
    _stop = (stop - datetime.datetime(1970, 1, 1)).total_seconds() * 1000
    data_id = 'user_generated_granule_{}'.format(granule_id)
    shape_arity = math.sqrt(len(gcps))
    shape = {'xArity': shape_arity,
             'yArity': shape_arity,
             'gcps': gcps}
    idf = {'uri': 'Python: {}'.format(data_id),
           'xArity': 0,
           'yArity': 0,
           'resolution': 1000000000,
           'subsampling_factor': 0,
           'shape': shape}
    metadata = {'sourceId': 1,
                'collectionId': collection_id,
                'granuleId': granule_id,
                'dataId': data_id,
                'dataModel': None,
                'start': int(_start),
                'stop': int(_stop),
                'uris': [idf],
                'title': data_id,
                'institution': '',
                'comment': '',
                'file_id': idf['uri'],
                'product_version': '0.0.0',
                'lat_min': min([gcp['lat'] for gcp in gcps]),
                'lat_max': max([gcp['lat'] for gcp in gcps]),
                'lon_min': min([gcp['lon'] for gcp in gcps]),
                'lon_max': max([gcp['lon'] for gcp in gcps]),
                'creator_email': '',
                'station_id': '',
                'platform': '',
                'sensor': ''}
    granule = {'id': {'granuleLevel': True,
                      'sourceId': 1,
                      'collectionId': collection_id,
                      'granuleId': granule_id,
                      'variableId': 0},
               'metadata': metadata,
               'data': {}}

    return (granule_id, granule)


def set_field(granule, field_name, field_data, dynamic_packing=True,
              as_float=False):
    """Attach data to a granule

    Parameters
    ----------
    granule : dict
        Dictionary that represents a granule, such as the output of
        :func:`SEAScope.lib.utils.create_granule`
    field_name : str
        Name that will identify the data within the granule. Please note that
        granules that belong to the same collection must *all* have the same
        fields
    field_data : numpy.ndarray or numpy.ma.MaskedArray
        Data that must be attached to the granule
    dynamic_packing : bool, optional
        If False the values contained in ``field_data`` will be included "as
        is", otherwise (default behavior) the min and max values of
        ``field_data`` are used to project the data in the ubyte domain: min
        value is mapped to 0, max value to 254 and intermediary values are
        interpolated. 255 is used to represent masked data.
    as_float : bool, optional
        If True data are converted to float32 values, otherwise they are
        converted to unsigned bytes (default behavior).

    Raises
    ------
    DynamicPackingFloatIncompatbilityError
        Raised when ``dynamic_packing`` and ``as_float`` are both True because
        dynamic packing is only implemented for unsigned bytes
    FieldShapeError
        Raised when ``field_data`` is neither 1D nor 2D
    IncompatibleFieldError
        Raised when a field cannot be added to a granule because it is not
        compatible with already attached fields (shape mismatch)
    """
    if (dynamic_packing is True) and (as_float is True):
        raise DynamicPackingFloatIncompatbilityError()

    import numpy
    x_arity = 0
    y_arity = 0
    shape = numpy.shape(field_data)
    dims = len(shape)
    if 1 == dims:
        x_arity = shape[0]
        y_arity = 1
    elif 2 == dims:
        # Invert x and y since 2D data is transposed
        x_arity = shape[1]
        y_arity = shape[0]
    else:
        raise FieldShapeError('Field data must be either 1D or 2D')

    if granule['metadata']['dataModel'] is None:
        if 1 == dims:
            granule['metadata']['dataModel'] = 'TIME'
            granule['metadata']['uris'][0]['shape']['xArity'] = x_arity
            granule['metadata']['uris'][0]['shape']['yArity'] = 1
        elif 2 == dims:
            granule['metadata']['dataModel'] = 'ROW_CELL'
        granule['metadata']['uris'][0]['xArity'] = x_arity
        granule['metadata']['uris'][0]['yArity'] = y_arity

    # Check compatibility between field and granule geometries
    if (2 == dims) and ('TIME' == granule['metadata']['dataModel']):
        raise IncompatibleFieldError('Adding a 2D field on a granule which '
                                     'already contains a 1D field is not '
                                     'supported')
    elif (1 == dims) and ('TIME' != granule['metadata']['dataModel']):
        raise IncompatibleFieldError('Adding a 1D field on a granule which '
                                     'already contains a 2D field is not '
                                     'supported')

    ref_x_arity = granule['metadata']['uris'][0]['xArity']
    ref_y_arity = granule['metadata']['uris'][0]['yArity']
    if (x_arity != ref_x_arity) or (y_arity != ref_y_arity):
        field_shape = '{}x{}'.format(x_arity, y_arity)
        ref_shape = '{}x{}'.format(ref_x_arity, ref_y_arity)
        raise IncompatibleFieldError('All fields must have the same shape '
                                     'within a granule: you can not add a '
                                     'field with a {} shape in a granule which'
                                     ' contains {} fields'.format(field_shape,
                                                                  ref_shape))

    # Find packing parameters
    offset = 0.0
    scale = 1.0
    if dynamic_packing:
        vmin = numpy.nanmin(field_data)
        vmax = numpy.nanmax(field_data)
        offset = vmin
        scale = (vmax - vmin) / 254
        if vmin == vmax:
            scale = 1.0

    # Data must be flipped and transposed for SEAScope
    if 2 == dims:
        transposed = numpy.transpose(field_data.copy())
    else:
        transposed = field_data[::-1]

    transposed_masked = numpy.ma.masked_invalid(transposed)
    transposed_mask = numpy.ma.getmaskarray(transposed_masked)
    mask_ind = numpy.where(transposed_mask)

    if as_float is False:
        # Pack data as uint8 (max value <=> 254, 255 is reserved for mask)
        transposed[mask_ind] = 0
        data_packed = (transposed - offset) / scale
        data_clipped = numpy.clip(data_packed, 0, 254).astype('uint8')

        # Mask invalid values (255 is used as fill value)
        fill_value = 255
        if numpy.any(transposed_mask):
            data_clipped[mask_ind] = fill_value

        data_type = 'uint8'
    else:
        transposed[mask_ind] = 0.0
        data_packed = (transposed - offset) / scale
        data_clipped = data_packed.astype('f4')

        # Mask invalid values using netCDF4 default fill value for float32
        fill_value = float.fromhex('0x1.e000000000000p+122')
        if numpy.any(transposed_mask):
            data_clipped[mask_ind] = fill_value

        data_type = 'float32'

    # Data must be stored as a 1D array (required for serialization)
    data_flattened = data_clipped.flatten('F')  # .tolist()

    channels = 1
    info = {'channels': channels,
            'xArity': x_arity,
            'yArity': y_arity,
            'dataType': data_type,
            'offsets': [offset],
            'scaleFactors': [scale],
            'fillValues': [fill_value]}

    granule['data'][field_name] = {'info': info,
                                   'buffer': data_flattened}


def get_lonlat(extraction, shape):
    """Rebuild lon/lat grids at an arbitrary resolution using interpolation,
    based on the GCPs provided by SEAScope extracted data.

    Parameters
    ----------
    extraction : dict
        Extraction from a single granule, deserialized and reformatted. This is
        usually *one* item of dictionary returned by
        :func:`SEAScope.lib.get_extracted_data`
    shape : tuple
        Shape of the resulting lon/lat grids

    Returns
    -------
    tuple(numpy.ndarray, numpy.ndarray)
        A tuple which contains two elements:

        - a grid of interpolated longitudes for the spatial covergae of the
          extraction

        - a grid of interpolated latitudes for the spatial coverage of the
          extraction

    Raises
    ------
    InvalidShapeError
        Error raised when the ``shape`` passed as parameter does not match a
        2D matrix

    Example
    -------
    >>> import SEAScope.lib
    >>> import SEAScope.lib.utils
    >>> extraction = SEAScope.lib.get_extracted_data('127.0.0.1', 11155)
    >>> print([str(x) for x in extraction.keys()])
    ['/seascope/data/ecmwf/ECMWF_20151201T12Z/ECMWF_20151201T12Z_idf_00.nc']
    >>> lons, lats = SEAScope.lib.utils.get_lonlat(extraction['/seascope/data/ecmwf/ECMWF_20151201T12Z/ECMWF_20151201T12Z_idf_00.nc'], (1000,1000))  # noqa:E501
    """
    import numpy

    if 2 != len(shape):
        raise InvalidShapeError('The "shape" parameter must have exactly two '
                                'components.')

    gcps = extraction['meta']['gcps']
    i_gcp = numpy.array([_['i'] for _ in gcps])
    j_gcp = numpy.array([_['j'] for _ in gcps])
    lon_gcp = numpy.array([_['lon'] for _ in gcps])
    lat_gcp = numpy.array([_['lat'] for _ in gcps])

    # Enforce longitude continuity (to be improved)
    if (lon_gcp[-1] - lon_gcp[0]) > 180.0:
        logger.warn('Difference between first and last longitude exceeds '
                    '180 degrees, assuming IDL crossing and remapping '
                    'longitudes in [0, 360]')
        lon_gcp = numpy.mod((lon_gcp + 360.0), 360.0)

    # Restore shape of the GCPs
    gcps_shape = (8, 8)  # hardcoded in SEAScope
    i_shaped = numpy.reshape(i_gcp, gcps_shape)
    j_shaped = numpy.reshape(j_gcp, gcps_shape)
    lon_shaped = numpy.reshape(lon_gcp, gcps_shape)
    lat_shaped = numpy.reshape(lat_gcp, gcps_shape)

    dst_lin = numpy.arange(0, shape[0])
    dst_pix = numpy.arange(0, shape[1])
    _dst_lin = numpy.tile(dst_lin[:, numpy.newaxis], (1, shape[1]))
    _dst_pix = numpy.tile(dst_pix[numpy.newaxis, :], (shape[0], 1))

    lon_2D, lat_2D = geoloc_from_gcps(lon_shaped, lat_shaped, j_shaped,
                                      i_shaped, _dst_lin, _dst_pix)

    return lon_2D, lat_2D


def geoloc_from_gcps(gcplon, gcplat, gcplin, gcppix, lin, pix):
    """"""
    import numpy
    import pyproj
    geod = pyproj.Geod(ellps='WGS84')
    fwd, bwd, dis = geod.inv(gcplon[:, :-1], gcplat[:, :-1],
                             gcplon[:, 1:], gcplat[:, 1:])

    # Find line and column for the top-left corner of the 4x4 GCPs cell which
    # contains the requested locations
    nlin, npix = gcplat.shape
    _gcplin = gcplin[:, 0]
    _gcppix = gcppix[0, :]
    top_line = numpy.searchsorted(_gcplin, lin, side='right') - 1
    left_column = numpy.searchsorted(_gcppix, pix, side='right') - 1

    # Make sure this line and column remain within the matrix and that there
    # are adjacent line and column to define the bottom-right corner of the 4x4
    # GCPs cell
    top_line = numpy.clip(top_line, 0, nlin - 2)
    bottom_line = top_line + 1
    left_column = numpy.clip(left_column, 0, npix - 2)
    right_column = left_column + 1

    # Compute coordinates of the requested locations in the 4x4 GCPs cell
    line_extent = _gcplin[bottom_line] - _gcplin[top_line]
    column_extent = _gcppix[right_column] - _gcppix[left_column]
    line_rel_pos = (lin - _gcplin[top_line]) / line_extent
    column_rel_pos = (pix - _gcppix[left_column]) / column_extent

    # Compute geographical coordinates of the requested locations projected on
    # the top and bottom lines
    lon1, lat1, _ = geod.fwd(gcplon[top_line, left_column],
                             gcplat[top_line, left_column],
                             fwd[top_line, left_column],
                             dis[top_line, left_column] * column_rel_pos)
    lon2, lat2, _ = geod.fwd(gcplon[bottom_line, left_column],
                             gcplat[bottom_line, left_column],
                             fwd[bottom_line, left_column],
                             dis[bottom_line, left_column] * column_rel_pos)

    # Compute the geographical coordinates of the requested locations projected
    # on a virtual column joining the projected points on the top and bottom
    # lines
    fwd12, bwd12, dis12 = geod.inv(lon1, lat1, lon2, lat2)
    lon, lat, _ = geod.fwd(lon1, lat1, fwd12, dis12 * line_rel_pos)

    return lon, lat


def import_granule_from_syntool(data_path, wkt, col_id, start, stop, field):
    """Read data extracted as Numpy objects using Syntool and import them in
    SEAScope.

    Note that Syntool dialog that pops up once the extraction is ready not
    only contains download links to get the extracted data but also the
    shape of the extracted area. You *must* save the shape (displayed in WKT
    format) in order to use this method as it is used as spatial coverage for
    the granule.

    Parameters
    ----------
    data_path : string
        Path of the data extracted in Numpy format
    wkt : string
        Shape of the extracted data, expressed in WKT format
    col_id : int
        Local, i.e. specific to the current data source, identifier for the
        collection that the granule containing the imported data will be
        attached to
    start: datetime.datetime
        Begining of the time coverage of the granule, in UTC (included)
    stop : datetime.datetime
        End of the time coverage of the granule, in UTC (excluded)
    field : string
        Name of the field that the extracted data corresponds to

    Example
    -------
    >>> import SEAScope.lib.utils
    >>> import datetime
    >>> coll_id, coll_dict = SEAScope.lib.utils.create_collection('Syntool data')  # noqa:E501
    >>> npy_path = '/tmp/3857_REMSS_MWOI_SST_v05.0-20190527120000-REMSS-L4_GHRSST-SSTfnd-MW_OI-GLOB-v02.0-fv05.0.npy'  # noqa:E501
    >>> shape_wkt = 'POLYGON((-42.3633 24.1267,-21.0938 24.1267,-21.0938 41.9677,-42.3633 41.9677,-42.3633 24.1267))'  # noqa:E501
    >>> start_dt = datetime.datetime(2019, 5, 27, 0, 0, 0)
    >>> stop_dt = datetime.datetime(2019, 5, 28, 0, 0, 0)
    >>> granule_id, granule_dict = SEAScope.lib.utils.import_granule_from_syntool(npy_path, shape_wkt, coll_id, start_dt, stop_dt, 'sst')  # noqa:E501
    >>> print(granule_id)
    1000
    >>> print(granule_dict)
    {'id': {'granuleLevel': True,
            'sourceId': 1,
            'collectionId': 11,
            'granuleId': 1000,
            'variableId': 0},
     'metadata': {'sourceId': 1,
                  'collectionId': 11,
                  'granuleId': 1000,
                  'dataId': 'user_generated_granule_1000',
                  'dataModel': 'ROW_CELL',
                  'start': 1558915200000,
                  'stop': 1559001600000,
                  'uris': [{'uri': 'Python: user_generated_granule_1000',
                            'xArity': 244,
                            'yArity': 242,
                            'resolution': 1000000000,
                            'subsampling_factor': 0,
                            'shape': {'xArity': 2.0,
                                      'yArity': 2.0,
                                      'gcps': [{'lon': -21.0938,
                                                'lat': 24.1267,
                                                'i': 243,
                                                'j': 241},
                                               {'lon': -42.3633,
                                                'lat': 24.1267,
                                                'i': 243,
                                                'j': 0},
                                               {'lon': -21.0938,
                                                'lat': 41.9677,
                                                'i': 0,
                                                'j': 241},
                                               {'lon': -42.3633,
                                                'lat': 41.9677,
                                                'i': 0,
                                                'j': 0}]}}],
                  'title': 'user_generated_granule_1000',
                  'institution': '',
                  'comment': '',
                  'file_id': 'Python: user_generated_granule_1000',
                  'product_version': '0.0.0',
                  'lat_min': 24.1267,
                  'lat_max': 41.9677,
                  'lon_min': -42.3633,
                  'lon_max': -21.0938,
                  'creator_email': '',
                  'station_id': '',
                  'platform': '',
                  'sensor': ''},
     'data': {'sst': {'info': {'channels': 1,
                               'xArity': 244,
                               'yArity': 242,
                               'dataType': 'uint8',
                               'offset': [0.0],
                               'scaleFactors': [1.0],
                               'fillValues': [255]},
                               'buffer': [...]}}}

    """
    import numpy
    data = numpy.load(data_path)
    data_transposed = numpy.ma.masked_values(numpy.transpose(data), 255)
    data_shape = numpy.shape(data_transposed)

    _wkt = wkt.strip()
    i0 = 1 + _wkt.rfind('(')
    i1 = _wkt.find(')')
    _vertices = [y.strip().split(' ') for y in _wkt[i0:i1].split(',')][:-1]
    vertices = [(float(x[0].strip()), float(x[1].strip())) for x in _vertices]

    # Build GCPs
    gcp0 = {'lon': vertices[1][0], 'lat': vertices[1][1],
            'i': data_shape[1] - 1, 'j': data_shape[0] - 1}
    gcp1 = {'lon': vertices[0][0], 'lat': vertices[0][1],
            'i': data_shape[1] - 1, 'j': 0}
    gcp2 = {'lon': vertices[2][0], 'lat': vertices[2][1],
            'i': 0, 'j': data_shape[0] - 1}
    gcp3 = {'lon': vertices[3][0], 'lat': vertices[3][1],
            'i': 0, 'j': 0}
    gcps = [gcp0, gcp1, gcp2, gcp3]

    granule_id, granule = create_granule(col_id, gcps, start, stop)
    set_field(granule, field, data_transposed, dynamic_packing=False)

    return (granule_id, granule)


def raw2data(raw_result):
    """Reformat raw data received from SEAScope, after it has been deserialized
    by FlatBuffers, into Numpy masked arrays.

    Parameters
    ----------
    raw_result : dict
        Deserialized version of the data received from SEAScope

    Returns
    -------
    dict
        Data received from SEAScope transformed into Numpy masked arrays
    """
    import numpy
    clean_data = {}
    for extraction in raw_result:
        uri = extraction['uri'].decode('utf-8')
        if uri not in clean_data:
            clean_data[uri] = {}
            shape = (extraction['yArity'], extraction['xArity'])
            clean_data[uri]['data'] = {}
            _start = datetime.datetime.utcfromtimestamp(extraction['start'])
            _stop = datetime.datetime.utcfromtimestamp(extraction['stop'])
            clean_data[uri]['meta'] = {"shape": shape,
                                       "gcps": extraction['gcps'],
                                       'fill_values': [],
                                       'fields': [],
                                       'start': _start,
                                       'stop': _stop}

        d = numpy.array(extraction['buffer'], dtype=numpy.float32)
        logger.debug('Full shape: {}'.format(d.shape))
        fill_values = extraction['fill_values']

        d0 = d[::4]
        data = [d0]

        if 1 < len(fill_values):
            d1 = d[1::4]
            data.append(d1)

        if 2 < len(fill_values):
            d2 = d[2::4]
            data.append(d2)

        is_mask = ((1 == len(data)) and (0 == len(extraction['fields'])) and
                   ('mask' == uri))
        for n in range(len(data)):
            if is_mask:
                field_name = 'mask'
                fill_value = 0
            else:
                field_name = (extraction['fields'][n]).decode('utf-8')
                fill_value = extraction['fill_values'][n]
            if field_name in clean_data[uri]['data']:
                # Field values already extracted from another variable
                continue
            dn_reshaped = numpy.reshape(data[n], shape)
            dn_masked = numpy.ma.masked_values(dn_reshaped, fill_value)
            dn_clean = numpy.ma.masked_invalid(dn_masked)
            clean_data[uri]['data'][field_name] = dn_clean
            clean_data[uri]['meta']['fill_values'].append(fill_value)
            clean_data[uri]['meta']['fields'].append(field_name)

    logger.info(' .:| Available granules |:.')
    logger.info(' --------------------------')
    for k in clean_data.keys():
        logger.info('• {}'.format(k))

    return clean_data



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: API

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Extraction(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsExtraction(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Extraction()
        x.Init(buf, n + offset)
        return x

    # Extraction
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Extraction
    def Id(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = o + self._tab.Pos
            from SEAScope.API.RenderableId import RenderableId
            obj = RenderableId()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Extraction
    def XArity(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # Extraction
    def YArity(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # Extraction
    def Channels(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # Extraction
    def Buffer(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Float32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # Extraction
    def BufferAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Float32Flags, o)
        return 0

    # Extraction
    def BufferLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Extraction
    def BufferIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0

    # Extraction
    def Shape(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 24
            from SEAScope.API.GCP import GCP
            obj = GCP()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Extraction
    def ShapeLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Extraction
    def ShapeIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        return o == 0

    # Extraction
    def Gcps(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 24
            from SEAScope.API.GCP import GCP
            obj = GCP()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Extraction
    def GcpsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Extraction
    def GcpsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        return o == 0

    # Extraction
    def FillValues(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Float32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # Extraction
    def FillValuesAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Float32Flags, o)
        return 0

    # Extraction
    def FillValuesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Extraction
    def FillValuesIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        return o == 0

    # Extraction
    def Uri(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Extraction
    def Fields(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # Extraction
    def FieldsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Extraction
    def FieldsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        return o == 0

    # Extraction
    def Start(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Extraction
    def Stop(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

def ExtractionStart(builder): builder.StartObject(12)
def ExtractionAddId(builder, id): builder.PrependStructSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(id), 0)
def ExtractionAddXArity(builder, xArity): builder.PrependUint32Slot(1, xArity, 0)
def ExtractionAddYArity(builder, yArity): builder.PrependUint32Slot(2, yArity, 0)
def ExtractionAddChannels(builder, channels): builder.PrependUint32Slot(3, channels, 0)
def ExtractionAddBuffer(builder, buffer): builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(buffer), 0)
def ExtractionStartBufferVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def ExtractionAddShape(builder, shape): builder.PrependUOffsetTRelativeSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(shape), 0)
def ExtractionStartShapeVector(builder, numElems): return builder.StartVector(24, numElems, 8)
def ExtractionAddGcps(builder, gcps): builder.PrependUOffsetTRelativeSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(gcps), 0)
def ExtractionStartGcpsVector(builder, numElems): return builder.StartVector(24, numElems, 8)
def ExtractionAddFillValues(builder, fillValues): builder.PrependUOffsetTRelativeSlot(7, flatbuffers.number_types.UOffsetTFlags.py_type(fillValues), 0)
def ExtractionStartFillValuesVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def ExtractionAddUri(builder, uri): builder.PrependUOffsetTRelativeSlot(8, flatbuffers.number_types.UOffsetTFlags.py_type(uri), 0)
def ExtractionAddFields(builder, fields): builder.PrependUOffsetTRelativeSlot(9, flatbuffers.number_types.UOffsetTFlags.py_type(fields), 0)
def ExtractionStartFieldsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def ExtractionAddStart(builder, start): builder.PrependUint64Slot(10, start, 0)
def ExtractionAddStop(builder, stop): builder.PrependUint64Slot(11, stop, 0)
def ExtractionEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: API

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class SelectVariableByLabel(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsSelectVariableByLabel(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = SelectVariableByLabel()
        x.Init(buf, n + offset)
        return x

    # SelectVariableByLabel
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # SelectVariableByLabel
    def CollectionLabel(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # SelectVariableByLabel
    def VariableLabel(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # SelectVariableByLabel
    def Selected(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # SelectVariableByLabel
    def Exclusive(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

def SelectVariableByLabelStart(builder): builder.StartObject(4)
def SelectVariableByLabelAddCollectionLabel(builder, collectionLabel): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(collectionLabel), 0)
def SelectVariableByLabelAddVariableLabel(builder, variableLabel): builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(variableLabel), 0)
def SelectVariableByLabelAddSelected(builder, selected): builder.PrependBoolSlot(2, selected, 0)
def SelectVariableByLabelAddExclusive(builder, exclusive): builder.PrependBoolSlot(3, exclusive, 0)
def SelectVariableByLabelEnd(builder): return builder.EndObject()



# vim: ts=4:sts=4:sw=4
#
# @author: <sylvain.herledan@oceandatalab.com>
# @date: 2016-09-08
#
# This file is part of SEAScope, a 3D visualisation and analysis application
# for satellite, in-situ and numerical model data.
#
# Copyright (C) 2014-2023 OceanDataLab
#
# SEAScope is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SEAScope. If not, see <https://www.gnu.org/licenses/>.

"""
This module handles the serialization of the getTimespan command and the
deserialization of the result sent by SEAScope
"""

import logging
import flatbuffers
import SEAScope.API.OpCode
import SEAScope.API.GetTimespanResponse
import SEAScope.API.Command
import SEAScope.types.timespan

logger = logging.getLogger(__name__)


def serialize(builder):
    """
    Serialize a getTimespan command using FlatBuffers.

    Parameters
    ----------
    builder : flatbuffers.builder.Builder
        The FlatBuffers builder instance which serializes data. If this
        parameter is None, then a new builder will be created

    Returns
    -------
    tuple(flatbuffers.builder.Builder, int)
        A tuple which contains two elements:

        - the :obj:`flatbuffers.builder.Builder` instance which has been used
          to serialize data
        - an :obj:`int` which is the address/offset of the serialized object
          in the builder buffer
    """
    if builder is None:
        builder = flatbuffers.Builder(0)

    opcode = SEAScope.API.OpCode.OpCode().getTimespan
    SEAScope.API.Command.CommandStart(builder)
    SEAScope.API.Command.CommandAddOpcode(builder, opcode)
    cmd = SEAScope.API.Command.CommandEnd(builder)

    return builder, cmd


def deserialize(buf):
    """
    Deserialize the response that SEAScope sends after executing the
    getTimespan command

    Parameters
    ----------
    buf : bytearray
        The buffer which contains the result of the getTimespan command
        serialized with FlatBuffers

    Returns
    -------
    dict
        The timespan currently selected in SEAScope. The dict is generated by
        the :func:`SEAScope.types.timespan.deserialize` method (see source code
        for more details)
    """
    cls = SEAScope.API.GetTimespanResponse.GetTimespanResponse
    res = cls.GetRootAsGetTimespanResponse(buf, 0)

    timespan = SEAScope.types.timespan.deserialize(res.Timespan())

    return timespan



# vim: ts=4:sts=4:sw=4
#
# @author: <sylvain.herledan@oceandatalab.com>
#
# This file is part of SEAScope, a 3D visualisation and analysis application
# for satellite, in-situ and numerical model data.
#
# Copyright (C) 2014-2023 OceanDataLab
#
# SEAScope is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SEAScope. If not, see <https://www.gnu.org/licenses/>.

__description__ = ('Python bindings to interact with the SEAScope application')
__version__ = '0.3'
__author__ = ('Sylvain Herlédan <sylvain.herledan@oceandatalab.com',)
__author_email__ = 'seascope@oceandatalab.com'
__url__ = 'https://seascope.oceandatalab.com/python.html'
__classifiers__ = ['Development Status :: 4 - Beta',
                   'Environment :: Console',
                   'Intended Audience :: Developers',
                   'Intended Audience :: Education',
                   'Intended Audience :: End Users/Desktop',
                   'Intended Audience :: Science/Research',
                   'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',  # noqa
                   'Operating System :: POSIX :: Linux',
                   'Operating System :: MacOS :: MacOS X',
                   'Operating System :: Microsoft :: Windows :: Windows 10',
                   'Operating System :: Microsoft :: Windows :: Windows 8.1',
                   'Operating System :: Microsoft :: Windows :: Windows 8',
                   'Operating System :: Microsoft :: Windows :: Windows 7',
                   'Programming Language :: Python :: 3.7',
                   'Topic :: Scientific/Engineering',
                   'Topic :: Scientific/Engineering :: Visualization',
                   'Topic :: Scientific/Engineering :: GIS']



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: API

class OpCode(object):
    addSource = 0
    removeSource = 1
    listSources = 2
    disconnect = 3
    addCollection = 4
    removeCollection = 5
    listCollections = 6
    addVariable = 7
    removeVariable = 8
    listVariables = 9
    setCurrentDatetime = 10
    getCurrentDatetime = 11
    addTimespan = 12
    removeTimespan = 13
    listTimespans = 14
    setTimespan = 15
    getTimespan = 16
    addTimestep = 17
    removeTimestep = 18
    listTimesteps = 19
    setTimestep = 20
    getTimestep = 21
    getNearestMode = 22
    setNearestMode = 23
    getDataAlwaysVisibleMode = 24
    setDataAlwaysVisibleMode = 25
    searchGranules = 26
    listVisibleData = 27
    setRendering = 28
    getRendering = 29
    getExtraction = 30
    addGranule = 31
    getCatalogues = 32
    resetCamera = 33
    zoomIn = 34
    zoomOut = 35
    setAltitude = 36
    lookAt = 37
    selectVariable = 38
    selectVariableByLabel = 39
    getVariableIdentifier = 40
    checkSchemaVersion = 41




# vim: ts=4:sts=4:sw=4
#
# @author: <sylvain.herledan@oceandatalab.com>
# @date: 2016-09-09
#
# This file is part of SEAScope, a 3D visualisation and analysis application
# for satellite, in-situ and numerical model data.
#
# Copyright (C) 2014-2023 OceanDataLab
#
# SEAScope is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SEAScope. If not, see <https://www.gnu.org/licenses/>.

"""
This module handles the serialization and deserialization of shape objects
"""

import logging
import flatbuffers
import SEAScope.API.Shape
import SEAScope.types.gcp

logger = logging.getLogger(__name__)


def serialize(builder, shp_obj):
    """
    Serialize a shape using FlatBuffers.

    Parameters
    ----------
    builder : flatbuffers.builder.Builder
        The FlatBuffers builder instance which serializes data. If this
        parameter is None, then a new builder will be created
    shp_obj : dict
        Dictionary which contains information about the shape to serialize
        It must have three keys:

        - ``xArity`` : an :obj:`int` which tells SEAScope how many columns it
          must allocate when reconstructing the GCPs matrix

        - ``yArity`` : an :obj:`int` which tells SEAScope how many rows it
          must allocate when reconstructing the GCPs matrix

        - ``gcps`` : a :obj:`list` of :obj:`dict` containing information
          about the GCPs of the shape.
          Each :obj:`dict` object must have all the keys required by
          :func:`SEAScope.types.gcp.serialize`

    Returns
    -------
    tuple(flatbuffers.builder.Builder, int)
        A tuple which contains two elements:

        - the :obj:`flatbuffers.builder.Builder` instance which has been used
          to serialize data
        - an :obj:`int` which is the address/offset of the serialized object
          in the builder buffer
    """
    if builder is None:
        builder = flatbuffers.Builder(0)

    x_arity = int(shp_obj['xArity'])
    y_arity = int(shp_obj['yArity'])
    gcps_len = len(shp_obj['gcps'])
    SEAScope.API.Shape.ShapeStartGcpsVector(builder, gcps_len)
    for gcp in shp_obj['gcps']:
        _, _gcp = SEAScope.types.gcp.serialize(builder, gcp)
    gcps = builder.EndVector(gcps_len)

    cls = SEAScope.API.Shape
    cls.ShapeStart(builder)
    cls.ShapeAddXArity(builder, x_arity)
    cls.ShapeAddYArity(builder, y_arity)
    cls.ShapeAddGcps(builder, gcps)
    shape = cls.ShapeEnd(builder)
    return builder, shape


def deserialize(o):
    """
    Rebuild a shape from a FlatBuffers buffer.

    Parameters
    ----------
    buf : bytearray
        The buffer which contains the shape object serialized with FlatBuffers

    Returns
    -------
    dict
        The deserialized shape object as a dictionary.
    """
    deserializer = SEAScope.types.gcp.deserialize
    s_obj = {}
    s_obj['xArity'] = o.XArity()
    s_obj['yArity'] = o.YArity()
    gcps_count = o.GcpsLength()
    s_obj['gcps'] = [deserializer(o.Gcps(x)) for x in range(0, gcps_count)]

    return s_obj



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: API

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class SetTimespan(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsSetTimespan(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = SetTimespan()
        x.Init(buf, n + offset)
        return x

    # SetTimespan
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # SetTimespan
    def Label(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

def SetTimespanStart(builder): builder.StartObject(1)
def SetTimespanAddLabel(builder, label): builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(label), 0)
def SetTimespanEnd(builder): return builder.EndObject()
